/*$Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*
 * pio.c -- i/o to and from processes
 */

#ifndef lint
static char rcsid[] = "$Header: pio.c 1.7 87/09/14 $";
#endif lint

#include <sys/errno.h>
#include "defs.h"
#include "main.h"
#include "machine.h"
#include "multi-p.h"
#include "library.h"
#include "keywords.h"

/*
 * These routines are used to access the debuggee process from
 * outside this module.
 *
 * They invoke "pio" which eventually leads to a call to "ptrace".
 * The system generates an I/O error when a ptrace fails.  During reads
 * these are ignored, during writes they are reported as an error, and
 * for anything else they cause a fatal error.
 */

#define cachehash(addr) ((unsigned) ((addr >> 2) % CSIZE))
#define WMASK	(~(sizeof(Word) - 1))

/*
 * These definitions are for the arguments to "pio".
 */

typedef enum { PREAD, PWRITE } PioOp;
typedef enum { TEXTSEG, DATASEG, UAREA } PioSeg;

private badaddr;
private read_err(), write_err();

/*
 * Read from the process user area
 */

public uread(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    INTFUNC *f;

    check_read(wp);
    f = onsyserr(EIO, read_err);
    badaddr = addr;
    if (wp->p_state == COREDUMP) {
	coredump_readuser(buff, addr, nbytes);
    } else {
	pio(PREAD, UAREA, buff, addr, nbytes);
    }
    (void) onsyserr(EIO, f);
}

/*
 * Read from the process' instruction area.
 */

public iread(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    INTFUNC *f;

    check_read(wp);
    f = onsyserr(EIO, read_err);
    badaddr = addr;
    if (wp->p_state == COREDUMP) {
	coredump_readtext(buff, addr, nbytes);
    } else {
	pio(PREAD, TEXTSEG, buff, addr, nbytes);
    }
    (void) onsyserr(EIO, f);
}

/* 
 * Write to the process' instruction area, usually in order to set
 * or unset a breakpoint.
 */

public iwrite(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    INTFUNC *f;

    check_write(wp);
    f = onsyserr(EIO, write_err);
    badaddr = addr;
    pio(PWRITE, TEXTSEG, buff, addr, nbytes);
    (void) onsyserr(EIO, f);
}

/*
 * Read for the process' data area.
 */

public dread(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    INTFUNC *f;

    check_read(wp);
    badaddr = addr;
    if (wp->p_state == COREDUMP) {
	f = onsyserr(EFAULT, read_err);
	coredump_readdata(buff, addr, nbytes);
	(void) onsyserr(EFAULT, f);
    } else {
	f = onsyserr(EIO, read_err);
	pio(PREAD, DATASEG, buff, addr, nbytes);
	(void) onsyserr(EIO, f);
    }
}

/*
 * Write to the process' data area.
 */

public dwrite(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    INTFUNC *f;

    check_write(wp);
    f = onsyserr(EIO, write_err);
    badaddr = addr;
    pio(PWRITE, DATASEG, buff, addr, nbytes);
    (void) onsyserr(EIO, f);
}

/*
 * Trap for errors in reading or writing to a process.
 * The current approach is to "ignore" read errors and complain
 * bitterly about write errors.
 */

private read_err()
{
    /*
     * Ignore.
     */
}

private write_err()
{
    error("can't write to process (address 0x%x)", badaddr);
}

/*
 * Structure for reading and writing by words, but dealing with bytes.
 */

typedef union {
    Word pword;
    Byte pbyte[sizeof(Word)];
} Pword;

/*
 * Read (write) from (to) the process' address space.
 * We must deal with ptrace's inability to look anywhere other
 * than at a word boundary.
 */

private Word fetch();
private store();

private pio(op, seg, buff, addr, nbytes)
PioOp op;
PioSeg seg;
char *buff;
Address addr;
int nbytes;
{
    register int i;
    register Address newaddr;
    register char *cp;
    char *bufend;
    Pword w;
    Address wordaddr;
    int byteoff;

#ifdef TRAPADDR
    Address low_addr, hi_addr;
    Node a1, a2;

    if (varIsSet("$checkpio")) {
	if (varIsSet("$pioaddrhi") and varIsSet("$pioaddrlow")) {
	    a1 = findvar(identname("$pioaddrhi", true));
	    a2 = findvar(identname("$pioaddrlow", true));
	    if (a1 == nil or a2 == nil or
		a1->op != O_LCON or a2->op != O_LCON)
		    low_addr = hi_addr = 0;
		else {
		    hi_addr = a1->value.lcon;
		    low_addr  = a2->value.lcon;
		}
	} else {
	    low_addr = hi_addr = 0;
	}
	if (addr >= low_addr and addr <= hi_addr) {
		trap_addr(op, seg, buff, addr, nbytes);
		return;
	}
    }
#endif TRAPADDR

    cp = buff;
    newaddr = addr;
    wordaddr = (newaddr&WMASK);
    if (wordaddr != newaddr) {
	w.pword = fetch(seg, wordaddr);
	for (i = newaddr - wordaddr; i < sizeof(Word) and nbytes > 0; i++) {
	    if (op == PREAD) {
		*cp++ = w.pbyte[i];
	    } else {
		w.pbyte[i] = *cp++;
	    }
	    nbytes--;
	}
	if (op == PWRITE) {
	    store(seg, wordaddr, w.pword);
	}
	newaddr = wordaddr + sizeof(Word);
    }
    byteoff = (nbytes&(~WMASK));
    nbytes -= byteoff;
    bufend = cp + nbytes;
    while (cp < bufend) {
	if (op == PREAD) {
	    *((Word *) cp) = fetch(seg, newaddr);
	} else {
	    store(seg, newaddr, *((Word *) cp));
	}
	cp += sizeof(Word);
	newaddr += sizeof(Word);
    }
    if (byteoff > 0) {
	w.pword = fetch(seg, newaddr);
	for (i = 0; i < byteoff; i++) {
	    if (op == PREAD) {
		*cp++ = w.pbyte[i];
	    } else {
		w.pbyte[i] = *cp++;
	    }
	}
	if (op == PWRITE) {
	    store(seg, newaddr, w.pword);
	}
    }
}

#ifdef TRAPADDR

/*
 * Used for debugging, this routine is entered when we access some
 * address in a range we desire to "trap".
 */

trap_addr(op, seg, buff, addr, nbytes)
PioOp op;
PioSeg seg;
char *buff;
Address addr;
int nbytes;
{
    printf("\nAddress trap:  address= 0x%x, nbytes= %d\n",
	addr, nbytes);
    printf("               op= ");
    switch (op) {
    case PREAD:
	printf("PREAD, seg= ");
	break;
    case PWRITE:
	printf("PWRITE, seg= ");
	break;
    default:
	printf("%d ??, seg= ", op);
    }
    switch (seg) {
    case UAREA:
	printf("UAREA\n");
	break;
    case TEXTSEG:
	printf("TEXTSEG\n");
	break;
    case DATASEG:
	printf("DATASEG\n");
	break;
    default:
	printf("%d ??\n", seg);
    }
    fflush(stdout);
}
#endif TRAPADDR

/*
 * Get a word from a process at the given address.
 * The address is assumed to be on a word boundary.
 *
 * A simple cache scheme is used to avoid redundant ptrace calls
 * to the instruction space since it is assumed to be pure.
 *
 * It is necessary to use a write-through scheme so that
 * breakpoints right next to each other don't interfere.
 */

private Integer nfetchs, nreads, nwrites;

private Word fetch(seg, addr)
PioSeg seg;
register int addr;
{
    register CacheWord *wordp;
    register Word w;

    switch (seg) {
	case TEXTSEG:
	    ++nfetchs;
	    wordp = &wp->p_tcache[cachehash(addr)];
	    if (addr == 0 or wordp->addr != addr) {
		++nreads;
		w = xptrace(PT_RTEXT, wp->p_pid, addr, 0);
		wordp->addr = addr;
		wordp->val = w;
	    } else {
		w = wordp->val;
	    }
	    break;

	case DATASEG:
	    w = xptrace(PT_RDATA, wp->p_pid, addr, 0);
	    break;

	case UAREA:
	    w = xptrace(PT_RUSER, wp->p_pid, addr, 0);
	    break;

	default:
	    panic("fetch: bad seg %d", seg);
	    /* NOTREACHED */
    }
    return w;
}

/*
 * Put a word into the process' address space at the given address.
 * The address is assumed to be on a word boundary.
 */

private store(seg, addr, data)
PioSeg seg;
int addr;
Word data;
{
    register CacheWord *wordp;

    switch (seg) {
	case TEXTSEG:
	    ++nwrites;
	    wordp = &wp->p_tcache[cachehash(addr)];
	    wordp->addr = addr;
	    wordp->val = data;
	    xptrace(PT_WTEXT, wp->p_pid, addr, data);
	    break;

	case DATASEG:
	    xptrace(PT_WDATA, wp->p_pid, addr, data);
	    break;

	default:
	    panic("store: bad seg %d", seg);
	    /* NOTREACHED */
    }
}

public printptraceinfo()
{
    printf("%d fetchs, %d reads, %d writes\n", nfetchs, nreads, nwrites);
}
