/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: queue.c 1.4 86/05/22 $";
#endif

/*
 * queue.c -- command serialization
 *
 * These routines deal with the problem of reentrancy caused by evaluation
 * of commands while in the SIGCHLD interrupt handler.  Many routines
 * (especially in libc) are not reentrant, and a call to eval() might call
 * any of them.  Since eval() must be called on behalf of routines that
 * are "running" for break and trace points, there is the potential that
 * non-reentrant routines might be reentered.  These routines guarantee
 * that the interrupt handler only calls eval() when it's "safe" to do so.
 *
 * the rules:
 *	- everything callable by babysitter() must be reentrant.
 *	- getinput() is the only proper way to get input from the user.
 *	- the only legal way to call eval from the interrupt side is to
 *	  call qcommand() and then doqueue().  doqueue() might end up
 *	  doing the command later.
 */

#include "defs.h"
#include "tree.h"
#include "main.h"
#include "multi-p.h"
#include "events.h"
#include "library.h"
#include <signal.h>

#define mask(b)	(1 << ((b) - 1))

struct queue {
	Cmdlist		q_cmd;
	int		q_id;
	struct queue	*q_next;
	int		q_eventid;
};

public boolean	debug_queue = false;	/* trace command queueing */
private struct queue *qhead = 0;
private struct queue *qtail = 0;


/*
 * get a line of input from 'f'.  just like fgets() except that it allows
 * the interrupt side to get things done while the read blocks.
 */
public char *getinput(s, n)
	char *s;
	int n;
{
	char *ret;
	int oldmask;

	resetproc();
	wwait();
	ret = fgets(s, n, win);
	setdefproc();

	return(ret);
}


public printq()
{
	register struct queue *q;
	int oldmask;

	oldmask = sigblock(mask(SIGCHLD));
	if (qhead == 0)
		fprintf(stderr, "Q: empty\n");
	for (q = qhead; q; q = q->q_next) {
		fprintf(stderr, "Q: 0x%x %%%d next=0x%x ", q, q->q_id, q->q_next);
		if (q == qhead)
			fprintf(stderr, "(head) ");
		if (q == qtail)
			fprintf(stderr, "(tail) ");
		fprintf(stderr, "\n");
	}
	(void) sigsetmask(oldmask);
}


/*
 * put a command on the queue.  SIGCHLD must be blocked.
 */
public qcommand(cmd, procid, eventid)
	Cmdlist cmd;
	int	procid;
	int	eventid;
{
	int oldmask;
	struct queue *q;

	if (debug_queue)
		fprintf(stderr, "queueing for %d\n", procid);
	q = (struct queue *)malloc(sizeof(struct queue));	/* XXX */
	q->q_next = (struct queue *)0;
	q->q_cmd = cmd;
	q->q_id = procid;
	q->q_eventid = eventid;
	if (qtail)
		qtail->q_next = q;
	qtail = q;
	if (qhead == 0)
		qhead = q;
	wp->p_pending = true;
}


/*
 * take a command off the queue and return it.  SIGCHLD must be blocked.
 */
private struct queue *deqcommand()
{
	int oldmask;
	struct queue *q;

	q = qhead;
	if (q) {
		if (q == q->q_next)
			error("q == q->q_next!!!");
		qhead = qhead->q_next;
		if (qhead == 0)
			qtail = 0;
		if (debug_queue)
			fprintf(stderr, "dequeueing for %d\n", q->q_id);
	}
	return(q);
}


/*
 * execute what's in the command queue if it's ok to do so.  SIGCHLD
 * and SIGINT must be blocked while this routine is executing.
 */
public doqueue()
{
	register struct queue *q, *qp;
	struct queue *contlist = 0;	/* list of procs to continue */
	Pstate prev;
	Lineno savecursrcline;
	String savecursource;
	String savecurpath;

	/*
	 * execute all the queued commands
	 */
	while( q = deqcommand() ) {
		if (debug_queue)
			fprintf(stderr, "executing for %d\n", q->q_id);

		/*
		 * create the environment for the process.  state of the
		 * process is set to stopped so we can do things to it.
		 */
		pushpid(q->q_id);
		wp->p_state = STOPPED;
		setjob(wp->p_job);
		savecurpath = curpath;
		savecursource = cursource;
		savecursrcline = cursrcline;
		setcurfunc(whatblock(reg(PROGCTR)));
		getsrcpos();
		cureventid = q->q_eventid;

		if (seterror() == 0) {

			/*
			 * execute the commands.
			 */
			prev = wp->p_newstate;
			evalcmdlist(q->q_cmd);
			fflush(stdout);
			unseterror();

		} else {

			/*
			 * there was an error in evalcmdlist()... recover.
			 */
			process_reset();
			pushpid(q->q_id);
			setstopped(BY_ERROR);
			wp->p_stopeventid = q->q_eventid;
		}
		
		/*
		 * get rid of the environment.  the logical thing to do would
		 * be to set the state back to PENDING, but the process will
		 * be handed to newstate() next, and it will make the correct
		 * state change.
		 */
		unsetjob();
		setsource(savecursource);
		setpath(savecurpath);
		cursrcline = savecursrcline;
		popproc();
		cureventid = 0;
		if (debug_queue)
			fprintf(stderr, "done executing for %d\n", q->q_id);

		/*
		 * add the process to the list of processes to continue,
		 * but don't add it if it's already there (don't want to
		 * continue a process more than once!).
		 */
		for (qp = contlist; qp && qp->q_id != q->q_id; qp = qp->q_next)
			;
		if (qp == 0) {
			q->q_next = contlist;
			contlist = q;
		} else
			free((char *)q);
	}

	/*
	 * "continue" all the processes.  this really means make them
	 * do a state transition either to RUNNING or STOPPED.
	 */
	for (qp = contlist; qp; qp = qp->q_next) {
		pushpid(qp->q_id);
		setjob(wp->p_job);
		wp->p_pending = false;
		newstate();
		free((char *)qp);
		unsetjob();
		popproc();
	}
}


/*
 * return when process p has stopped.  the process state will be
 * changed by the SIGCHLD handler when the process stops, so this
 * routine just pauses until the process's status becomes STOPPED.
 */
public waitfor(p)
	Process *p;
{
	int oldmask;

	oldmask = sigblock(mask(SIGCHLD));		/* lock */
	doqueue();
	while (p->p_state == RUNNING || p->p_state == PENDING) {
		(void) sigpause(oldmask);		/* pause */
		doqueue();
	}
	(void) sigsetmask(oldmask);			/* unlock */
}
