/*LINTLIBRARY*/

/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: shbrk.c 1.20 87/05/28 $";
#endif

/*
 * shbrk.c
 *	Parallel run-time basic shared memory allocator.
 *
 * TODO:
 *	When mmap() handles "extended" u_mmap[]'s, can remove the EMFILE code.
 *	Worry about taking signal in sbrk()/shmbrk() (use process local
 *		signal stack; only really issue for multi-tasker).
 *	Add support for shrinking?
 */

/* $Log:	shbrk.c,v $
 */

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/signal.h>
#include <machine/vmparam.h>

#include <parallel/parallel.h>
#include "parc.h"

/*
 * Module local variables.
 */

private	int _priv_stack_limit;			/* size limit for priv stack */
private	static	char	*myshbrk;		/* process local max sbrk */
shared	static	char	*curshbrk;		/* global for all processes */
shared	static	slock_t	shbrk_lock = L_UNLOCKED;

char *	_shared_end;

extern _stack_incr;
extern _stack_hole;
extern _stack_limit;
/*
 * round an address down to the beginning of a STACK_INCR boundary
 */
#define STACK_ROUND(a)	(int *)((int)(a) & (~(_stack_incr - 1)))
#define STACK_ROUND_UP(a) (int *)((int)(a + _stack_incr) & (~(_stack_incr - 1)))

/*
 * stack_top and stack_base bound the shared parent stack.  Require that
 *	stack pointer be in the interval:
 *
 *		[stack_base,stack_top)
 *	
 */
private static int	*stack_top;		/* shared stack top address */
private static int	*stack_base;		/* shared stack lower limit */
/*
 * This is the lowest stack address which has been mapped by the current
 * process.  This is of interest only to child processes.
 */
private static int	*stack_mybottom;		 

/*
 * This is the lowest stack address that the parent address has allocated.
 * The parent can dynamicaly grow the stack down to stack_base, and this 
 * growth would be reflected in stack_curbottom.  But, child processes would
 * still have to map this new growth, as reflected in stack_mybottom.
 */
shared static int	*stack_curbottom;	/* current stack bottom */

/*
 * _init_shbrk()
 *	Initialize basic shared-memory allocator.
 *
 * Called from _ppinit().
 */

static int	segvh();
static struct sigvec vec = 
			{segvh, (sigmask(SIGSEGV)|sigmask(SIGCHLD)), 1};
private static int segvstack[512];
private static struct sigstack sigstruct = {(caddr_t)&segvstack[511], 0};
int sigstack();
extern private int _single_stack;
extern char *getenv();

_init_shbrk()
{
	/*extern	shared char _shend_; */
	extern int _do_shared_stack;

	/*
	 * Set up SIGSEGV handler; catch signal and grow shared-data
	 * and shared stack if appropriate.
	 */

	(void)sigstack(&sigstruct, (struct sigstack *)0);
	(void)sigvec(SIGSEGV, &vec, (struct sigvec *)0);

	/*
	 * Initialize state variables.
	 */

	if (_do_shared_stack)
		curshbrk = myshbrk = ((char *)stack_top + _stack_hole);
	else
		/*curshbrk = myshbrk = &_shend_;*/
		curshbrk = myshbrk = _shared_end;
	if (curshbrk >= ((char *)MAXADDR - _priv_stack_limit - _stack_hole)) {
		printf("ppinit: shared break address too big\n");
		errno = ENOEXEC;
		return(-1);
	}
	return(0);
}

/*
 * shsbrk()
 *	shared-memory sbrk(); create new shared space, return start address.
 */

char *
shsbrk(incr)
	int	incr;
{
	char	*val;
	register int sig_mask;
	register slock_t *spin_ptr = &shbrk_lock;

	if (incr == 0)
		return(curshbrk);	/* note: can race with others! */

	if (incr < 0) {			/* can't shrink */
		errno = EINVAL;
		return((char *) -1);
	}

	/*
	 * Lock data-file to insure one shared-brk at a time (and
	 * stable value of curshbrk).
	 *
	 * Note that if we're running on the shared stack, the other
	 * processes are "idle".  So, there's no need to block the
	 * SIGSEGV and SIGCHILD signals, because the child processes
	 * shouldn't be dying.  Note also that we CAN't block these
	 * signals here because we need SIGSEGV to potentially grow
	 * the stack we're running on.
	 *
	 * But, if we're NOT running on the shared stack, there IS a 
	 * need to block these signals.  If a sibling process gets a
	 * "real" SIGSEGV, and the parent re-sends this SIGSEGV to
	 * a process which is holding the lock below, deadlocks can
	 * occur!!
	 * 
	 */

	if (!_single_stack)
		sig_mask = sigblock(sigmask(SIGSEGV)|sigmask(SIGCHLD));
	S_LOCK(spin_ptr);

	val = curshbrk;
	if (lshbrk(val + incr) < 0)
		val = (char *) -1;

	S_UNLOCK(spin_ptr);
	if (!_single_stack)
		(void) sigsetmask(sig_mask);

	return(val);
}

/*
 * shbrk()
 *	Add shared address space, analogous to brk().
 *
 * Just front-end to lshbrk() -- lock file and try to grow.
 */

shbrk(newbrk)
	char	*newbrk;
{
	int	val;
	register int sig_mask;
	register slock_t *spin_ptr = &shbrk_lock;

	if (!_single_stack)
		sig_mask = sigblock(sigmask(SIGSEGV)|sigmask(SIGCHLD));
	S_LOCK(spin_ptr);

	val = lshbrk(newbrk);

	S_UNLOCK(spin_ptr);
	if (!_single_stack)
		(void) sigsetmask(sig_mask);

	return(val);
}

/*
 * lshbrk()
 *	"Locked" shbrk().  Internal procedure.
 *
 * Assumes caller locked shmem file, to serialize access to this code.
 */

static
lshbrk(newbrk)
	register char	*newbrk;
{
	register char	*cur;			/* actual brk (rounded) */
	register char	*new;			/* requested brk (rounded) */
	extern	shared char	_shstart_;	/* start of shared stuff */
#ifndef CONCAT
	static	char	*lastmmap = &_shstart_;	/* for EMFILE recovery */
#endif CONCAT

	/*
	 * Not allowed to shrink self (yet).
	 *
	 * Also, check for overflow with bottom of stack.
	 */

	if ((newbrk < myshbrk) || (newbrk > ((char *)MAXADDR - 
					_priv_stack_limit - _stack_hole))) {
		errno = EINVAL;
		return(-1);
	}

	/*
	 * Growing: if no actual increase (rounding/etc), just adjust
	 * (self) myshbrk.
	 */

	cur = PGRND(myshbrk);
	if (cur >= newbrk) {
		myshbrk = newbrk;
		if (newbrk > curshbrk)
			curshbrk = myshbrk;
		return(0);
	}
	new = PGRND(newbrk);

	/*
	 * Need to mmap() more of file.  If get EMFILE from mmap(),
	 * need to re-map last piece.  Should arrange mmap() to extend
	 * when appropriate to avoid EMFILE errors.
	 */

	if (MMAP(_shm_fd, cur, new - cur, cur - &_shstart_) < 0) {
#ifdef CONCAT
		errno = ENOMEM;
		return(-1);
#else	CONCAT
		if (errno == EMFILE) {
			/*
			 * Make room by deleting last map.
			 */
			(void) munmap(lastmmap, cur - lastmmap);
			if (MMAP(_shm_fd, lastmmap, new - lastmmap, 
						lastmmap - &_shstart_) < 0) {
/*
				if (MMAP(_shm_fd, lastmmap, cur - lastmmap,
						lastmmap - &_shstart_) < 0) {
					perror("shbrk error");
				}
*/
				errno = ENOMEM;
				return(-1);
			}
			cur = lastmmap;
		} else {
			errno = ENOMEM;
			return(-1);
		}
#endif	CONCAT
	}

#ifndef CONCAT
	lastmmap = cur;			/* for later re-map */
#endif CONCAT
	myshbrk = newbrk;		/* maybe not aligned */
	if (newbrk > curshbrk)
		curshbrk = newbrk;	/* grew the global memory */

	return(0);			/* success! */
}

/*
 * _init_shstk()
 *	Initialize basic shared-memory allocator.
 *
 * Called from _ppinit().
 *	return pointer to the bottom of the shared stack.  This
 * 	holds the stack context for the EXIT, RET instructions 
 *	out of ppinit().
 */

int *
_init_shstk(ctxt_bottom, ctxt_top)
int *ctxt_bottom;
int *ctxt_top;
{
	int copy_len;
	int *new_sp;

	extern int _stk_fd;
	/*extern	shared char _shend_;*/

	/*
	 * initialize stack bounds
	 */
	stack_base = STACK_ROUND_UP(_shared_end + _stack_hole);
	/*stack_base = STACK_ROUND_UP(&_shend_ + _stack_hole); */
	stack_curbottom = stack_mybottom = stack_top = stack_base + 
						_stack_limit/sizeof(int);
	/*
	 * check for overflow into private stack area
	 * By definition at this point, there's no possibiltiy for overflow
	 * into shared data region.
	 */
	if (stack_top > STACK_ROUND(MAXADDR - ((_priv_stack_limit + _stack_hole)/(sizeof (int))))) {
		printf("not enough address space for stack\n");
		errno = ENOMEM;
		return((int *)-1);
	}

	/*
	 * actually allocate some stack segment space.
	 */
	if (grow_stack(STACK_ROUND((stack_top - 1)))) {
#ifdef DEBUG
		printf("%x %x\n", STACK_ROUND((stack_top - 1)), stack_top);
		printf("grow_stack failed\n");
#endif DEBUG
		errno = ENOEXEC;
		return((int *)-1);
	}
	stack_curbottom = stack_mybottom;

	/*
	 * now, copy context from the private stack onto the shared stack
	 * want to copy ONLY the PROCEDURE CALL context for ppinit() that 
         * will be left on the current stack after this function exits.
	 * This is also the place we want the new stack pointer to be when 
	 * ppinit() switches to the shared stack. 
	 */

	copy_len = (char *)ctxt_top - (char *)ctxt_bottom;
	new_sp = (stack_top) - (copy_len/(sizeof (int)));
#ifdef DEBUG
	printf("copy_len %x new_sp %x ctxt_top %x ctxt_bottom %x\n",
		copy_len, new_sp, ctxt_top, ctxt_bottom);
#endif DEBUG
	bcopy((char *)(ctxt_bottom), (char *)new_sp, (unsigned int)copy_len);
	return(new_sp);
}

/*
 * grow_stack()
 *	process grows its view of the shared stack.
 *
 * 	This function is called from init_shstk() and segvh().  
 */

static
grow_stack(new_bottom)
register int *new_bottom;
{
	register int size;
	register int offset;

	size = (stack_mybottom - new_bottom)*sizeof(int);
	offset = (new_bottom - stack_base)*sizeof(int);

#ifdef DEBUG
	printf("new_bottom %x size %x offset %x\n", new_bottom, size, offset);
#endif DEBUG
	if (MMAP(_stk_fd, (caddr_t)new_bottom, size, offset) == -1) {
		errno = ENOMEM;
		return(-1);
	}
	stack_mybottom = new_bottom;
	return(0);
}

/*
 * segvh()
 *	SIGSEGV handler.
 *
 * Check for growth.  If growth succeeded, just return.  Otherwise, replace
 * handler and retry (this should core-dump).
 *
 */
/*ARGSUSED*/
static
segvh(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
	if(_check_segv(sig, code, scp) == 0)
		(void) signal(SIGSEGV, SIG_DFL);
}
/*
 * Check to see if the segv results in growth, or is a failure.
 *	0 return means failure.
 *	1 return means growth occurred.
 *
 * This is meant to be called when a SIGSEGV has occurred, and further
 * SIGSEGV's are blocked.  Also, this should be invoked on a process-private
 * stack.
 *
 * add code to deal with mapping/growing the shared stack area.
 */

/*ARGSUSED*/
_check_segv(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
	register slock_t *spin_ptr = &shbrk_lock;
	register int *fa = (int *)code;
	extern int _single_stack;
	extern int _do_shared_stack;

#ifdef DEBUG
	printf("check_segv fa %x\n", code);
#endif DEBUG
	/*
	 * a sig code of -1 indicates the signal originated
	 * from a kill() system call, rather than a real program
	 * fault
	 */
	if (fa == (int *)-1) {
		(void) kill(getpid(), SIGSEGV);
		return(0);
	}


	/*
 	 * Since only the parent process can GROW the stack, 
	 * since ONLY the parent process will ever MODIFY the shared variable 
	 * "stack_curbottom". Therefore, no locks are needed on stack_bottom.
	 *
	 * The variable "stack_mybottom is private to each process, so no locks
	 * needed there either.
	 */

	/*
	 * Are we using the shared stack mechanism?
	 */
	if (_do_shared_stack) {
		/*
		 * Are we running in parallel at this time?
		 *
		 *	if not, then the shared stack is still dynamically
		 *	growable.  Otherwise, the shared stack is not growable,
		 *	but we may still need to map parts of the shared stack.
		 */
		if (_single_stack)  {
			if ( (fa >= stack_base) && (fa < stack_curbottom)) {
				/*
			  	 * GROW the shared stack
			 	 */
#ifdef DEBUG
				printf("Grow stack\n");
#endif DEBUG
				if (grow_stack(STACK_ROUND(fa)) == 0) {
					stack_curbottom = stack_mybottom;
					return(1);
				}
			} else {
				/*
				 * have we overflowed our stack?
				 * if so, we won't be able to unwind out
				 * of the current signal handler.  So,
				 * we should deliver a new signal on
				 * our current stack.  This signal
				 * handler should then kill off the kids.
				 */
				if ((int *)scp->sc_sp < stack_base) {
					printf("limits stacksize exceeded\n");
					(void) kill(getpid(), SIGILL);
				}
			}
		} else if ((fa >= stack_curbottom) && (fa < stack_mybottom)) {
#ifdef DEBUG
			printf("map stack\n");
#endif DEBUG
			/*
			 * just map MORE of the shared stack into my address space
			 */
			if (grow_stack(stack_curbottom) == 0) {
				return(1);
			}
		}
	}
		

	/*
	 * Fall through and see if the shared heap needs growing
	 *
	 * If process doesn't have all mapped, map it and retry.
	 * Otherwise turn off SIGSEGV handling.
	 *
	 * Lock data-file during this to stabilize curshbrk and
	 * insure only one lshbrk() at a time.
	 *
	 * Return retries instruction, will die for real if turn off
	 * signal handling here.
	 */

	S_LOCK(spin_ptr);

	if (myshbrk >= curshbrk || lshbrk(curshbrk) < 0) {
		(void) signal(SIGSEGV, SIG_DFL);
		S_UNLOCK(spin_ptr);
		return(0);
	}
	S_UNLOCK(spin_ptr);
	return(1);
}
